@@ -1,4 +1,4 @@
-// Copyright © 2015-2019 Hilko Bengen <bengen@hilluzination.de>
+// Copyright © 2015-2020 Hilko Bengen <bengen@hilluzination.de>
 // All rights reserved.
 //
 // Use of this source code is governed by the license that can be
@@ -16,17 +16,21 @@ package yara
 
 #include <yara.h>
 
-void compilerCallback(int, char*, int, char*, void*);
+void compilerCallback(int, char*, int, YR_RULE*, char*, void*);
+char* includeCallback(char*, char*, char*, void*);
+void freeCallback(char*, void*);
 */
 import "C"
 import (
 	"errors"
+	"os"
+	"reflect"
 	"runtime"
 	"unsafe"
 )
 
 //export compilerCallback
-func compilerCallback(errorLevel C.int, filename *C.char, linenumber C.int, message *C.char, userData unsafe.Pointer) {
+func compilerCallback(errorLevel C.int, filename *C.char, linenumber C.int, rule *C.YR_RULE, message *C.char, userData unsafe.Pointer) {
 	c := callbackData.Get(userData).(*Compiler)
 	msg := CompilerMessage{
 		Filename: C.GoString(filename),
@@ -44,16 +48,15 @@ func compilerCallback(errorLevel C.int, filename *C.char, linenumber C.int, mess
 // A Compiler encapsulates the YARA compiler that transforms rules
 // into YARA's internal, binary form which in turn is used for
 // scanning files or memory blocks.
+//
+// Since this type contains a C pointer to a YR_COMPILER structure
+// that may be automatically freed, it should not be copied.
 type Compiler struct {
-	*compiler
 	Errors   []CompilerMessage
 	Warnings []CompilerMessage
 	// used for include callback
 	callbackData unsafe.Pointer
-}
-
-type compiler struct {
-	cptr *C.YR_COMPILER
+	cptr         *C.YR_COMPILER
 }
 
 // A CompilerMessage contains an error or warning message produced
@@ -70,13 +73,19 @@ func NewCompiler() (*Compiler, error) {
 	if err := newError(C.yr_compiler_create(&yrCompiler)); err != nil {
 		return nil, err
 	}
-	c := &Compiler{compiler: &compiler{cptr: yrCompiler}}
-	runtime.SetFinalizer(c.compiler, (*compiler).finalize)
+	c := &Compiler{cptr: yrCompiler}
+	runtime.SetFinalizer(c, (*Compiler).Destroy)
 	return c, nil
 }
 
-func (c *compiler) finalize() {
-	C.yr_compiler_destroy(c.cptr)
+// Destroy destroys the YARA data structure representing a compiler.
+//
+// It should not be necessary to call this method directly.
+func (c *Compiler) Destroy() {
+	if c.cptr != nil {
+		C.yr_compiler_destroy(c.cptr)
+		c.cptr = nil
+	}
 	runtime.SetFinalizer(c, nil)
 }
 
@@ -87,16 +96,34 @@ func (c *Compiler) setCallbackData(ptr unsafe.Pointer) {
 	c.callbackData = ptr
 }
 
-// Destroy destroys the YARA data structure representing a compiler.
-// Since a Finalizer for the underlying YR_COMPILER structure is
-// automatically set up on creation, it should not be necessary to
-// explicitly call this method.
-func (c *Compiler) Destroy() {
-	c.setCallbackData(nil)
-	if c.compiler != nil {
-		c.compiler.finalize()
-		c.compiler = nil
+// AddFile compiles rules from a file. Rules are added to the
+// specified namespace.
+//
+// If this function returns an error, the Compiler object will become
+// unusable.
+func (c *Compiler) AddFile(file *os.File, namespace string) (err error) {
+	if c.cptr.errors != 0 {
+		return errors.New("Compiler cannot be used after parse error")
 	}
+	var ns *C.char
+	if namespace != "" {
+		ns = C.CString(namespace)
+		defer C.free(unsafe.Pointer(ns))
+	}
+	filename := C.CString(file.Name())
+	defer C.free(unsafe.Pointer(filename))
+	id := callbackData.Put(c)
+	defer callbackData.Delete(id)
+	C.yr_compiler_set_callback(c.cptr, C.YR_COMPILER_CALLBACK_FUNC(C.compilerCallback), id)
+	numErrors := int(C.yr_compiler_add_fd(c.cptr, (C.YR_FILE_DESCRIPTOR)(file.Fd()), ns, filename))
+	if numErrors > 0 {
+		var buf [1024]C.char
+		msg := C.GoString(C.yr_compiler_get_error_message(
+			c.cptr, (*C.char)(unsafe.Pointer(&buf[0])), 1024))
+		err = errors.New(msg)
+	}
+	runtime.KeepAlive(c)
+	return
 }
 
 // AddString compiles rules from a string. Rules are added to the
@@ -125,7 +152,7 @@ func (c *Compiler) AddString(rules string, namespace string) (err error) {
 			c.cptr, (*C.char)(unsafe.Pointer(&buf[0])), 1024))
 		err = errors.New(msg)
 	}
-	keepAlive(c)
+	runtime.KeepAlive(c)
 	return
 }
 
@@ -157,7 +184,7 @@ func (c *Compiler) DefineVariable(identifier string, value interface{}) (err err
 	default:
 		err = errors.New("wrong value type passed to DefineVariable; bool, int64, float64, string are accepted")
 	}
-	keepAlive(c)
+	runtime.KeepAlive(c)
 	return
 }
 
@@ -170,12 +197,82 @@ func (c *Compiler) GetRules() (*Rules, error) {
 	if err := newError(C.yr_compiler_get_rules(c.cptr, &yrRules)); err != nil {
 		return nil, err
 	}
-	r := &Rules{rules: &rules{cptr: yrRules}}
-	runtime.SetFinalizer(r.rules, (*rules).finalize)
-	keepAlive(c)
+	r := &Rules{cptr: yrRules}
+	runtime.SetFinalizer(r, (*Rules).Destroy)
+	runtime.KeepAlive(c)
 	return r, nil
 }
 
+//export includeCallback
+func includeCallback(name, filename, namespace *C.char, userData unsafe.Pointer) *C.char {
+	callbackFunc := callbackData.Get(userData).(CompilerIncludeFunc)
+	if buf := callbackFunc(
+		C.GoString(name), C.GoString(filename), C.GoString(namespace),
+	); buf != nil {
+		ptr := C.calloc(1, C.size_t(len(buf)+1))
+		if ptr == nil {
+			return nil
+		}
+		outbuf := make([]byte, 0)
+		hdr := (*reflect.SliceHeader)(unsafe.Pointer(&outbuf))
+		hdr.Data, hdr.Len = uintptr(ptr), len(buf)+1
+		copy(outbuf, buf)
+		return (*C.char)(ptr)
+	}
+	return nil
+}
+
+//export freeCallback
+func freeCallback(callbackResultPtr *C.char, userData unsafe.Pointer) {
+	if callbackResultPtr != nil {
+		C.free(unsafe.Pointer(callbackResultPtr))
+	}
+	return
+}
+
+// CompilerIncludeFunc is the type of the function that can be
+// registered through SetIncludeCallback. It is called for every
+// include statement encountered by the compiler. The argument "name"
+// specifies the rule file to be included, "filename" specifies the
+// name of the rule file where the include statement has been
+// encountered, and "namespace" specifies the rule namespace. The sole
+// return value is a byte slice containing the contents of the
+// included file. A return value of nil signals an error to the YARA
+// compiler.
+//
+// See also: yr_compiler_set_include_callback in the YARA C API
+// documentation.
+type CompilerIncludeFunc func(name, filename, namespace string) []byte
+
+// SetIncludeCallback sets up cb as an include callback that is called
+// (through Go glue code) by the YARA compiler for every include
+// statement.
+func (c *Compiler) SetIncludeCallback(cb CompilerIncludeFunc) {
+	if cb == nil {
+		c.DisableIncludes()
+		return
+	}
+	id := callbackData.Put(cb)
+	c.setCallbackData(id)
+	C.yr_compiler_set_include_callback(
+		c.cptr,
+		C.YR_COMPILER_INCLUDE_CALLBACK_FUNC(C.includeCallback),
+		C.YR_COMPILER_INCLUDE_FREE_FUNC(C.freeCallback),
+		id,
+	)
+	runtime.KeepAlive(c)
+	return
+}
+
+// DisableIncludes disables all include statements in the compiler.
+// See yr_compiler_set_include_callbacks.
+func (c *Compiler) DisableIncludes() {
+	C.yr_compiler_set_include_callback(c.cptr, nil, nil, nil)
+	c.setCallbackData(nil)
+	runtime.KeepAlive(c)
+	return
+}
+
 // Compile compiles rules and an (optional) set of variables into a
 // Rules object in a single step.
 func Compile(rules string, variables map[string]interface{}) (r *Rules, err error) {
